-- Copyright   :  (c) Vidas Daudaravicius 2012
-- License     :  BSD3-style (see LICENSE)
-- Version     : 0.1

module Context where

import Files
import Data.List (group,sort,span,sortBy,isInfixOf) 
import System.Directory
import MakeDict
import Ngram
import Dendogram
import Data.Function
import qualified System.IO as IO
import qualified Data.ByteString.Char8 as BS
import qualified Data.ByteString as B
import qualified Data.Map as M
import DeepSeq
import qualified Data.List as DL
import Data.Word        
import Data.Bits        ((.|.),(.&.),shiftL,shiftR)
import Data.Char        (chr,ord)



partOfSegment hIn word = do
                    end <- IO.hIsEOF hIn
                    if end 
                       then return []
                       else do
                              ln <- BS.hGetLine hIn
                              seg <- return $ BS.takeWhile (/='\t') ln
                              seg2 <- return $ BS.concat [BS.pack "_",seg,BS.pack "_"]
                              if BS.isInfixOf word seg2
                                 then do
                                        ls <- partOfSegment hIn word
                                        return $ seg:ls
                                 else partOfSegment hIn word


partOfToken fin word = do
                    end <- IO.hIsEOF fin
                    if end 
                       then return []
                       else do
                              ln <- BS.hGetLine fin
                              seg <- return $ BS.takeWhile (/='\t') ln
                              if BS.isInfixOf word seg
                                 then do
                                        ls <- partOfToken fin word
                                        return $ seg:ls
                                 else partOfToken fin word

getIdfDict :: IO.Handle -> BS.ByteString -> IO (BS.ByteString,[(BS.ByteString,Double)])
getIdfDict hIdf word = do
                            end <- IO.hIsEOF hIdf
                            if end
                               then return (word,[])
                               else do
                                    ln <- BS.hGetLine hIdf
                                    (w:dict) <- return $ splitOn '\t' ln
                                    if w < word
                                       then getIdfDict hIdf word
                                       else if word == w
                                               then return $ (word,Dendogram.parseIndex dict)
                                               else return (word,[])
                                       
                                    
                           

--fromString :: String -> B.ByteString
fromString xs = B.pack (encodeU xs)


-- | Encode a single Haskell Char to a list of Word8 values, in UTF8 format.
encodeChar :: Char -> [Word8]
encodeChar = map fromIntegral . go . ord
 where
  go oc
   | oc <= 0x7f       = [oc]

   | oc <= 0x7ff      = [ 0xc0 + (oc `shiftR` 6)
                        , 0x80 + oc .&. 0x3f
                        ]

   | oc <= 0xffff     = [ 0xe0 + (oc `shiftR` 12)
                        , 0x80 + ((oc `shiftR` 6) .&. 0x3f)
                        , 0x80 + oc .&. 0x3f
                        ]
   | otherwise        = [ 0xf0 + (oc `shiftR` 18)
                        , 0x80 + ((oc `shiftR` 12) .&. 0x3f)
                        , 0x80 + ((oc `shiftR` 6) .&. 0x3f)
                        , 0x80 + oc .&. 0x3f
                        ]


-- | Encode a Haskell String to a list of Word8 values, in UTF8 format.
encodeU :: String -> [Word8]
encodeU = concatMap encodeChar




look indF idfFile word cand = do 
           hContext <- IO.openFile idfFile IO.ReadMode
--           (w,seg) <- getIdfDict hContext (fromString word)
           (w,seg) <- getIdfDict hContext (BS.pack word)
           avg <- return $ (sum (map (snd) seg))/ ((toEnum (length seg))::Double)
           IO.hClose hContext
           putStrLn "looking for candidates"
           fs <- getCandidateFile indF $ map (fst) $ filter (\(_,i) -> i>avg) seg
--           fs <- getCandidateFile indF $ map (fst) $ filter ((>avg).snd) seg
           putStrLn $ "Found candidates:" ++ (show $ length fs)
           candidates <- return $ sort $ map (fst) $ fs -- take cand fs
--           BS.putStrLn $ BS.unlines $ map ( \(f,fr) -> (BS.concat [f,BS.singleton ' ',(BS.pack $ show fr)])) $ take 50 fs
           hContext <- IO.openFile idfFile IO.ReadMode
           ds' <- countContextCosines candidates seg hContext True BS.empty []
           IO.hClose hContext
           let lst = (sortBy (compare  `on` snd) ds')
           let list = drop ((length lst)-cand) lst
           BS.putStrLn $ BS.unlines [ BS.concat[f,BS.singleton ' ',(BS.pack $ show sim)]  |  (f,sim) <- list]



countContextCosines [] _ _ _ _ rez  = return rez
countContextCosines wss@(word:ws) dict hContext next ln rez' = do
                          end <- IO.hIsEOF hContext
                          let rez = if length rez' > 500 then (take 200 $ sortBy (\ (_,x) (_,y) -> compare y x) rez') else rez'
                          if end
                             then return rez
                             else do
                                    txt <- if next
                                                then BS.hGetLine hContext
                                                else return ln
                                    (w:dict') <- return $ splitOn '\t' txt
                                    if w < word
                                       then do --BS.putStrLn w
                                               countContextCosines wss dict hContext True txt $! rez
                                       else if word == w
                                               then do
                                                     cos' <- return $! cosine 0.00001 dict $! Dendogram.parseIndex dict'
                                                     countContextCosines ws dict hContext True txt $! seq cos' ((w,(cos')):rez)
                                               else do
                                                     countContextCosines ws dict hContext False txt $! rez


--countCosines d dir [] = return []
--countCosines d dir (f:fs) = do
--                          cs <- countCosines d dir fs
--                          txt <- BS.readFile $ (dir++(BS.unpack f)++".tfidf")
--                          cos' <- return $ cosine 0.00001 d $! map (\ln ->   let (w,tf) = BS.span (/='\t') ln in (w, (read (BS.unpack tf)::Double))) $ tail $ BS.lines txt
--                          return $! ((f,(cos')):cs)



makeWordContext dir dir2 ext d1 span fInd fIdf = do
                                    files <- getFiles [dir] ext
                                    fs2 <- wordContext span files
                                    [ws] <- mergeWordContext fs2
                                    splitIntoFiles' ws (ws++"'")
                                    removeFile ws
                                    renameFile (ws++"'") ws
                                    putStrLn "Done: merge word context"
                                    makeContextIDF (ws) (ws++"'")
                                    putStrLn "done: IDF prepare"
                                    mergeContextDicts (ws++"'") d1
                                    putStrLn "done: ContextDictionary"
                                    docs  <- readDictDocs d1
                                    uniD' <- readDictUTF8 d1
                                    uniD  <- return $!! M.fromAscList uniD'
                                    countContextTFIDF uniD docs (ws++"'") fIdf 
                                    putStrLn "done .tfidf dictionaries"
                                    removeFile $ ws
                                    removeFile $ ws++"'"
                                    builContextIndex (fIdf ++ "_") fInd
                                    removeFile (fIdf ++ "_")

avgContextLength file = do
                         text <- BS.readFile file
                         let tLines = BS.lines text
                         return $! toEnum(DL.foldl' (\ len line -> len + (length $ splitOn '\t' line))  0 tLines) / (toEnum (length tLines))

tabBS = BS.pack "\t"
numBS x = BS.pack $! show x

printConLine2 hOut word ((k,v):xs) = do
                                 BS.hPutStr hOut $! BS.concat [tabBS,k,tabBS, word,tabBS,numBS v]
                                 printConLine2 hOut word xs
printConLine2 hOut word [] = return ()



countContextTFIDF' avgDocLen dictD docs hIn hOut hOut2 = do
                                          end <- IO.hIsEOF hIn
                                          if end
                                             then return ()
                                             else do
                                                   ln1 <- BS.hGetLine hIn
                                                   (word, d) <- return $ parseDict $ splitOn '\t' ln1
                                                   let d1 = map (\(k,v) -> (k,toEnum v::Double)) d
                                                   let docLen = DL.foldl' (\fr1 (_,fr2) -> fr1 + fr2 ) 0 $! d1
                                                   let idf   = seq docLen [ (k,tf_idf)   | (k,v)<-d1, let tf_idf = tfidfNorm (M.findWithDefault 1 k dictD) docLen avgDocLen v docs , tf_idf>0.01 ]
                                                   let idf'' = map (snd) $! idf
                                                   let len   = toEnum (length (idf''))::Double
                                                   let avg   = (sum idf'') / (2 * len)
                                                   let idf''''  = seq avg filter (\(k,v) -> v>=avg) idf
                                                   let idf2  = if (length idf'' > 200) then (sort $! take 200  $! sortBy (compare `on` snd) idf'''') else idf'''' 
                                                   BS.hPutStr hOut word
                                                   mapM_ (\ (k,v) -> BS.hPutStr hOut  $ BS.concat [tabBS,k,tabBS,numBS v]) $! idf
--                                                   mapM_ (\ (k,v) -> BS.hPutStr hOut2 $ BS.concat [tabBS,k,tabBS, word,tabBS,numBS v]) $! idf2
                                                   printConLine2 hOut2 word $! idf2
                                                   if (not.null) idf  then BS.hPutStrLn hOut BS.empty  else return ()
                                                   if (not.null) idf2 then BS.hPutStrLn hOut2 BS.empty else return ()
                                                   countContextTFIDF' avgDocLen dictD docs hIn hOut hOut2


countContextTFIDF dictD docs fIn fOut = do
                                  avgDocLen <- avgContextLength fIn
                                  hIn <- IO.openFile fIn IO.ReadMode
                                  hOut <- IO.openFile fOut IO.WriteMode
                                  hOut2 <- IO.openFile (fOut++"_") IO.WriteMode
                                  countContextTFIDF' avgDocLen dictD docs hIn hOut hOut2
                                  IO.hClose hIn
                                  IO.hClose hOut
                                  IO.hClose hOut2



mergeContextDicts' hIn hOut d dd = do
                   end <- IO.hIsEOF hIn
                   if end
                      then do
                             d'' <- return $!! mergeContextSorted d dd
                             BS.hPutStrLn hOut (BS.pack $ show $ length d'')
                             mapM_ (\(w,f) -> BS.hPutStrLn hOut $ BS.concat [w,BS.pack "\t",BS.pack (show f)]) $! d''
                      else do
                            ln1 <- BS.hGetLine hIn
                            d1 <- return $!! parseDict2 $ tail $ splitOn '\t' ln1
                            d' <- return $!! mergeContextSorted d1 dd
                            if length d' > 5000
                               then do
                                     d'' <- return $!! mergeContextSorted d d'
                                     mergeContextDicts'  hIn hOut d'' [] 
                               else do
                                     mergeContextDicts'  hIn hOut d d' 


mergeContextSorted xs [] = xs
mergeContextSorted [] ys = ys
mergeContextSorted xss@((x,f1):xs) yss@((y,f2):ys) 
            | x == y = (x,f1+f2):(mergeContextSorted xs ys)
            | x < y  = (x,f1):(mergeContextSorted xs yss)
            | otherwise = (y,f2):(mergeContextSorted xss ys)


                           

mergeContextDicts inF outF = do
                          hIn  <- IO.openFile inF IO.ReadMode
                          end <- IO.hIsEOF hIn
                          if end
                             then do
                                 IO.hClose hIn
                                 return ()
                             else do
                                   hOut <- IO.openFile outF IO.WriteMode
                                   mergeContextDicts' hIn hOut [] []
                                   IO.hClose hOut
                                   IO.hClose hIn
                                             
                                             

parseDict2 []       = []
parseDict2 (w:f:ws) = (w,(read (BS.unpack f)::Int)):(parseDict2 ws)
parseDict (x:xs) = (x,parseDict2 xs)

makeContextIDF inF outF = do
                          hIn  <- IO.openFile inF IO.ReadMode
                          hOut <- IO.openFile outF IO.WriteMode
                          makeContextIDF' hIn hOut
                          IO.hClose hIn
                          IO.hClose hOut

makeContextIDF' hIn hOut = do
                   end <- IO.hIsEOF hIn
                   if end
                      then return ()
                      else do
                       ln <- BS.hGetLine hIn
                       (word,list) <- return $ parseDict $ splitOn '\t' ln
                       BS.hPutStr hOut $ word
                       mapM_ (\(ln,f) -> BS.hPutStr hOut $ BS.concat [BS.pack "\t", ln, BS.pack "\t1"] ) list 
                       BS.hPutStrLn hOut BS.empty
                       makeContextIDF' hIn hOut



splitOn c ln = BS.lines $ BS.map (\ x -> if x==c then '\n' else x) $ ln



parseDictIndex' []       = []
parseDictIndex' (w:f:idf:ws) = (w,BS.concat [f,BS.pack "\t",idf]):(parseDictIndex' ws)


mergeContextIndex' hIn fOut fC c d dd = do
                   end <- IO.hIsEOF hIn
                   if end
                      then do
                             hOut <- IO.openFile (fOut ++ (show fC) ++ ".c_") IO.WriteMode
                             mapM_ (\(w,f) -> BS.hPutStrLn hOut $ BS.concat [w,BS.pack "\t",f]) $! mergeContextSortedIndex d dd
                             IO.hClose hOut
                             return fC
                      else do
                            (fC2,d2, dd2) <- if length d > 60000
                                              then do
                                                    hOut <- IO.openFile (fOut ++ (show fC) ++ ".c_") IO.WriteMode
                                                    mapM_ (\(w,f) -> BS.hPutStrLn hOut $ BS.concat [w,BS.pack "\t",f]) $! mergeContextSortedIndex d dd
                                                    IO.hClose hOut
                                                    return (fC+1,[],[])
                                              else return (fC,d,dd)
                            ln1 <- BS.hGetLine hIn
                            d1 <- return $ parseDictIndex' $ tail $ splitOn '\t' ln1
                            d' <- return $ mergeContextSortedIndex d1 dd2
                            if length d' > 5000
                               then do
                                     ddd <- return $ force $ mergeContextSortedIndex d2 d'
                                     mergeContextIndex'  hIn fOut fC2 (c+1) ddd [] 
                               else do
                                     ddd <- return $ force d'
                                     mergeContextIndex'  hIn fOut fC2 (c+1) d2 ddd 


mergeContextSortedIndex xs [] = xs
mergeContextSortedIndex [] ys = ys
mergeContextSortedIndex xss@((x,f1):xs) yss@((y,f2):ys) 
            | x == y = (force(x,BS.concat [f1,BS.pack "\t",f2])):(mergeContextSortedIndex xs ys)
            | x < y  = (force(x,f1)):(mergeContextSortedIndex xs yss)
            | otherwise = (force(y,f2)):(mergeContextSortedIndex xss ys)


                           

builContextIndex inF outF = do
                          hIn  <- IO.openFile inF IO.ReadMode
                          end <- IO.hIsEOF hIn
                          if end
                             then do
                                 IO.hClose hIn
                                 return ()
                             else do
                                   cs <- mergeContextIndex' hIn outF 1 1 [] []
                                   IO.hClose hIn
                                   mergeConIndF outF cs

mergeConIndF outF cs = do
                     [f] <- mergeIndF2 [outF++ (show c) ++ ".c_" | c <- [1..cs]]
                     renameFile f outF

mergeIndF2 [] = return []                          
mergeIndF2 [f] = return [f]                          
mergeIndF2 fs = do
                  fs2 <- mergeIndFT fs
                  mergeIndF2 fs2


mergeIndFT []  = return []
mergeIndFT [f] = return [f]
mergeIndFT (f1:f2:fs) = do
                          fin1 <- IO.openFile f1 IO.ReadMode
                          fin2 <- IO.openFile f2 IO.ReadMode
                          fout <- IO.openFile (f1++"_")  IO.WriteMode
                          mergeIndContextF fin1 BS.empty fin2 BS.empty fout
                          IO.hClose fout
                          IO.hClose fin1
                          IO.hClose fin2
                          removeFile f1
                          removeFile f2
                          renameFile (f1++"_") f1
                          fs2  <- mergeIndFT fs
                          return (f1:fs2)


mergeIndContextF::IO.Handle->BS.ByteString->IO.Handle->BS.ByteString->IO.Handle->IO()
mergeIndContextF fin1 s1 fin2 s2 fout =
                               do
                                 end1 <- IO.hIsEOF fin1
                                 end2 <- IO.hIsEOF fin2
                                 x <- if (BS.null s1 && not end1) then BS.hGetLine fin1 else return s1  
                                 y <- if (BS.null s2 && not end2) then BS.hGetLine fin2 else return s2  
                                 if (BS.null x && BS.null y) 
                                   then return ()
                                   else if (BS.null x)
                                      then do
                                           BS.hPutStrLn fout y
                                           mergeIndContextF fin1 x fin2 BS.empty fout
                                      else if (BS.null y)
                                           then do
                                                BS.hPutStrLn fout x
                                                mergeIndContextF fin1 BS.empty fin2 y fout
                                           else do
                                              (w1,ind1) <- return $ BS.break (== '\t') x
                                              (w2,ind2) <- return $ BS.break (== '\t') y
                                              if (w1 == w2) 
                                                then do
                                                      BS.hPutStr fout $ w1
                                                      BS.hPutStr fout ind1 
                                                      BS.hPutStrLn fout ind2
                                                      mergeIndContextF fin1 BS.empty fin2 BS.empty fout
                                                else if (w1<w2)
                                                      then do
                                                            BS.hPutStrLn fout x
                                                            mergeIndContextF fin1 BS.empty fin2 y fout
                                                      else do
                                                            BS.hPutStrLn fout y
                                                            mergeIndContextF fin1 x fin2 BS.empty fout


wordContext span (fileName:files) = do
                            text <- BS.readFile fileName
                            BS.writeFile (fileName++".con") $! BS.unlines $! map (\x -> BS.concat [(head x),BS.pack "\t",BS.pack (show $ length x)]) $! group $! sort $! getBigrams span $ BS.words $ text
                            ls <- wordContext span files
                            return $ (fileName++".con"):ls
wordContext span [] = return []



splitIntoFiles' file file' = do
                          hIn <- IO.openFile file IO.ReadMode
                          hOut <- IO.openFile file' IO.WriteMode
                          splitIntoTokenContextFiles' BS.empty BS.empty hIn hOut
                          IO.hClose hIn
                          IO.hClose hOut


splitIntoTokenContextFiles' token context hIn hOut
                 = do 
                      if not $ BS.null token 
                        then BS.hPutStr hOut $ BS.concat[BS.pack "\t", context]
                        else return ()
                      end <- IO.hIsEOF hIn
                      if end
                         then return ()
                         else do
                                 ln <- BS.hGetLine hIn
                                 (token2,context2) <- return $ BS.span (/=' ') $ ln
                                 if token2 == token
                                    then return ()
                                    else do
                                          if not (BS.null token) 
                                             then BS.hPutStrLn hOut BS.empty
                                             else return ()
                                          BS.hPutStr hOut $ (token2)
                                 splitIntoTokenContextFiles' token2 (BS.tail context2) hIn hOut


getBigrams span [] = []
getBigrams span (w:ws) = [ BS.concat[w,BS.pack " ",y] | y <- sp ] ++ [ BS.concat[y,BS.pack " ",w] | y <- sp ] ++ getBigrams span ws
                          where 
                              sp = take span ws


mergeWordContextT [] = return []
mergeWordContextT [f]= return [f]
mergeWordContextT (f1:f2:fs) = do
                          fin1 <- IO.openFile f1 IO.ReadMode
                          fin2 <- IO.openFile f2 IO.ReadMode
                          fout <- IO.openFile (f1++"_")  IO.WriteMode
                          mergeWordContextF fin1 BS.empty fin2 BS.empty fout
                          IO.hClose fout
                          IO.hClose fin1
                          IO.hClose fin2
                          removeFile f1
                          removeFile f2
                          fs2  <- mergeWordContextT fs
                          return ((f1++"_"):fs2)

mergeWordContext [] = return []
mergeWordContext [fs] = return [fs]
mergeWordContext fs = do
                        fs2 <- mergeWordContextT fs
                        mergeWordContext fs2

mergeWordContextF::IO.Handle->BS.ByteString->IO.Handle->BS.ByteString->IO.Handle->IO()
mergeWordContextF fin1 s1 fin2 s2 fout =
                               do
                                 end1 <- IO.hIsEOF fin1
                                 end2 <- IO.hIsEOF fin2
                                 x <- if (BS.null s1 && not end1) then BS.hGetLine fin1 else return s1  
                                 y <- if (BS.null s2 && not end2) then BS.hGetLine fin2 else return s2  
                                 if (BS.null x && BS.null y) 
                                   then return ()
                                   else if (BS.null x)
                                      then do
                                           BS.hPutStrLn fout y
                                           mergeWordContextF fin1 x fin2 BS.empty fout
                                      else if (BS.null y)
                                           then do
                                                BS.hPutStrLn fout x
                                                mergeWordContextF fin1 BS.empty fin2 y fout
                                           else do
                                              (w1,fr1) <- return $ BS.break (== '\t') x
                                              (w2,fr2) <- return $ BS.break (== '\t') y
                                              if (w1 == w2) 
                                                then do
                                                      BS.hPutStrLn fout $ BS.concat [w1,BS.pack "\t",BS.pack (show((read (BS.unpack fr1)::Int) + (read (BS.unpack fr2)::Int)))]
                                                      mergeWordContextF fin1 BS.empty fin2 BS.empty fout
                                                else if (w1<w2)
                                                      then do
                                                            BS.hPutStrLn fout x
                                                            mergeWordContextF fin1 BS.empty fin2 y fout
                                                      else do
                                                            BS.hPutStrLn fout y
                                                            mergeWordContextF fin1 x fin2 BS.empty fout


