-- Copyright   :  (c) Vidas Daudaravicius 2012
-- License     :  BSD3-style (see LICENSE)
-- Version     : 0.1


module Dendogram where

import qualified System.IO as IO
import qualified Data.Map as M
import qualified Data.List as L
import Data.Function
import Segmentation
import Ngram
import MakeDict
import DeepSeq
import qualified Data.ByteString.Char8 as BS


getDict file ext = do
                  dict <- readDictUTF8 (file++ext)
                  return (file,dict)
--lookForDocument :: String -> Int -> String -> Double -> IO ()
lookForDocument file ngram threshold classified limit = do
           segF <- case ngram of
                       "seg" ->  preprocesSegFile file threshold classified
                       "bi"  ->  preprocesBiFile file classified
                       "uni" ->  preprocesUniFile file classified
           seg  <- readDictUTF8 segF
           putStrLn "looking for candidate files"
           fs <-   case ngram of
                       "seg" ->  getCandidateFile "ind.seg" $ map (fst) seg
                       "bi"  ->  getCandidateFile "ind.bi"  $ map (fst) seg
                       "uni" ->  getCandidateFile "ind.uni" $ map (fst) seg
           putStrLn $ "Found candidates:" ++ (show $ length fs)
--           putStrLn $ show $ take 50 fs
           candidates <- return $ map (fst) $ take 500 fs -- take count fs
           ds <-  case ngram of
                       "seg" ->  sequence ( map (\file -> getDict (BS.unpack file) ".txt.tok.seg.freq.tfidf") candidates )
                       "bi"  ->  sequence ( map (\file -> getDict (BS.unpack file) ".txt.tok.bi.freq.tfidf" )  candidates )
                       "uni" ->  sequence ( map (\file -> getDict (BS.unpack file) ".txt.tok.uni.freq.tfidf") candidates )
--           ds    <- return (zip candidates ds')
           list  <- return (L.sortBy (compare  `on` snd) [ (f1,(cosine 0.01 seg d1))  | (f1,d1)<-ds])
           let res = [ f ++ " " ++ show sim  |  (f,sim) <- list]
           putStrLn $ unlines $ drop ((length res) - limit)  res


--preprocesSegFile :: String -> Double -> IO String
preprocesSegFile fileName threshold classified = do
                        tokeniseFile fileName classified
                        uniF <- readDocument $ fileName++".tok"
                        writeDict (fileName ++ ".tok.uni.freq") uniF 1
                        uniF  <- readDict'' $ fileName++".tok.uni.freq"
                        uni <- readDict' "dict.uni" $ uniF
                        biF <- readDocumentBi $ fileName++".tok"
                        writeDictBi (fileName++".tok.bi.freq") biF 1
                        biF  <- readDict'' $ fileName++".tok.bi.freq"
                        bi  <- readDict' "dict.bi" biF
                        segmentDocument uni bi 100000 threshold $ fileName++".tok"
                        segF  <- readDocument $ fileName++".tok.seg"
                        writeDict (fileName++".tok.seg.freq") segF 1
                        segD  <- readDict' "dict.segD" $!! M.toList segF
                        docs  <- readDictDocs "dict.uniD"
                        avgDocLength <- readFile $ "seg.avglen"
                        countTFIDF segD docs (read avgDocLength) ".tfidf" $ fileName++".tok.seg.freq"
                        return (fileName++".tok.seg.freq.tfidf")

--preprocesUniFile :: String -> IO String
preprocesUniFile fileName classified = do
                        tokeniseFile fileName classified
                        uniF <- readDocument $ fileName++".tok"
                        writeDict (fileName++".tok.uni.freq") uniF 1
                        uniD  <- readDict' "dict.uniD" $! M.toList uniF
                        docs  <- readDictDocs "dict.uniD"
                        avgDocLength <- readFile $ "uni.avglen"
                        countTFIDF uniD docs (read avgDocLength) ".tfidf" $ fileName++".tok.uni.freq"
                        return (fileName++".tok.uni.freq.tfidf")

--preprocesBiFile :: String -> IO String
preprocesBiFile fileName classified = do
                        tokeniseFile fileName classified
                        biF <- readDocumentBi $ fileName++".tok"
                        writeDictBi (fileName++".tok.bi.freq") biF 1
                        biF  <- readDict'' $ fileName++".tok.bi.freq"
                        biD  <- readDict' "dict.biD" $! biF
                        docs  <- readDictDocs "dict.biD"
                        avgDocLength <- readFile $ "bi.avglen"
                        countTFIDF biD docs (read avgDocLength) ".tfidf" $ fileName++".tok.bi.freq"
                        return (fileName++".tok.bi.freq.tfidf")


mergeSorted xs [] = xs
mergeSorted [] ys = ys
mergeSorted xss@((x,f1):xs) yss@((y,f2):ys) 
            | x == y = (x,force (f1+f2)):(mergeSorted xs ys)
            | x < y  = (x,f1):(mergeSorted xs yss)
            | otherwise = (y,f2):(mergeSorted xss ys)


splitFiles  ln = BS.lines $ BS.map (\ x -> if x=='\t' then '\n' else x) $ BS.drop 1 ln


parseIndex []       = []
parseIndex (w:idf:ws) = (w,read (BS.unpack idf)::Double):(parseIndex ws)


getMatching [] = []
getMatching ((x,fs):xs) =  mergeSorted (L.sort $! parseIndex $ splitFiles fs) $! (getMatching xs)
--getMatching ((x,fs):xs) =  mergeSorted (map (\x -> (x,1)) $ L.sort $ splitFiles fs) $ (getMatching xs)

--filterSorted :: IO.Handle -> [BS.ByteString] -> IO [(BS.ByteString,BS.ByteString)]
filterSorted hIn next ln' [] = return []
filterSorted hIn next ln' yss@(y:ys) = do
                         end <- IO.hIsEOF hIn
                         if end
                           then return []
                           else do
                                 ln <- if next
                                             then BS.hGetLine hIn
                                             else return ln'
                                 let (w,fs) = (BS.span (/='\t') ln)
                                 if w==y
                                   then do
                                          fs2 <- filterSorted hIn True BS.empty yss
                                          return $ (w,fs):fs2
                                   else if w < y
                                          then filterSorted hIn True BS.empty yss
                                          else filterSorted hIn False ln ys


getCandidateFile :: String -> [BS.ByteString] -> IO [(BS.ByteString,Double)]
getCandidateFile indFile ws = do
                               ind <- IO.openFile indFile IO.ReadMode
                               ls <- filterSorted ind True BS.empty ws
                               IO.hClose ind
                               return $ L.sortBy (\(x,f1) (y,f2) -> compare f2 f1) $ getMatching ls

---------------------------------------------------------

zipDict []                []        disc        = []
zipDict []                ys        disc        = [ (disc,f) | (w,f) <- ys]
zipDict xs                []        disc        = [ (f,disc) | (w,f) <- xs]
zipDict allx@((w1,f1):xs) ally@((w2,f2):ys) disc
             | w1 == w2  = (f1,f2): zipDict xs ys disc
             | w1 < w2   = (f1,disc): zipDict xs ally disc
             | otherwise = (disc,f2): zipDict allx ys disc

---------------------------------------------------------
--(Ord a, Floating a1) => [(a, a1)] -> [(a, a1)] -> a1
cosine disc d1 d2
      | null d1 || null d2 = 0
      | x2 /= 0 && x3 /= 0 = x1 / ( (sqrt x2) * (sqrt x3) )
      | otherwise          = 0
           where
              x1 = L.foldl' (\x (f1,f2) -> f1*f2+x) 0 $! d
              x2 = L.foldl' (\x (f1,f2) -> f1^2+x)  0 $! d
              x3 = L.foldl' (\x (f1,f2) -> f2^2+x)  0 $! d
              d = zipDict d1 d2 disc

fn file = drop 1 $ dropWhile (/='/')  $ takeWhile (/='.') file
    