-- Copyright   :  (c) Vidas Daudaravicius 2012
-- License     :  BSD3-style (see LICENSE)
-- Version     : 0.1

module MakeDict where

import qualified System.IO as IO
import System.Directory
import qualified Data.Map as M
import Files
import Ngram
import Segmentation
import DeepSeq
import qualified Data.ByteString.Char8 as BS
import qualified Data.List as DL
import Data.Function
import Control.Monad (foldM)


avgDocLen files = do
                   lSize <- foldM (\len f -> do
                                               d <- readDict'' f
                                               return $! DL.foldl' (\x (_,y) -> x + y) len d
                                  ) 
                                  0 
                                  files
                   return $ lSize / toEnum(length files)

countDirTFIDF dir = do
        docs  <- readDictDocs $ "dict.uniD"
        uniD'  <- readDict'' $ "dict.uniD"
        uniD <- return $!! M.fromAscList uniD'
        files <- getFiles [dir] ".uni.freq"
        avgDocLength <- avgDocLen files
        writeFile ("uni.avglen") $ show avgDocLength
        mapM_ (countTFIDF uniD docs avgDocLength ".tfidf") files
        putStrLn "Done: uni.freq.tfidf dictionaries"
        biD'   <- readDict'' $ "dict.biD"
        biD <- return $!! M.fromAscList biD'
        files <- getFiles [dir] ".bi.freq"
        avgDocLength <- avgDocLen files
        writeFile ("bi.avglen") $ show avgDocLength
        mapM_ (countTFIDF biD docs avgDocLength ".tfidf") files
        putStrLn "Done: bi.freq.tfidf dictionaries"
        segD'  <- readDict'' $ "dict.segD"
        segD <- return $!! M.fromAscList segD'
        files <- getFiles [dir] ".seg.freq"
        avgDocLength <- avgDocLen files
        writeFile ("seg.avglen") $ show avgDocLength
        mapM_ (countTFIDF segD docs avgDocLength ".tfidf") files
        putStrLn "Done: seg.freq.tfidf dictionaries"


countSegmentationDirTFIDF dir = do
        docs  <- readDictDocs $ "dict.uniD"
        segD'  <- readDict'' $ "dict.segD"
        segD <- return $!! M.fromAscList segD'
        files <- getFiles [dir] ".seg.freq"
        avgDocLength <- avgDocLen files
        writeFile ("seg.avglen") $ show avgDocLength
        mapM_ (countTFIDF segD docs avgDocLength ".tfidf") files
        putStrLn "Done: seg.freq.tfidf dictionaries"




countTFIDF dic docs avgDocLen ext file = do
        docD'<- readDictUTF8 (file)
        docLen <- return $ DL.foldl' (\fr1 (_,fr2) -> fr1 + fr2 ) 0 docD'
        docD <- return $!! M.fromAscList docD'
        uni' <- return ( M.filter ( > 0) $! M.mapWithKey (\ k v -> tfidfNorm (M.findWithDefault 1 k dic) docLen avgDocLen v docs ) docD)
        writeDict (file++ext) uni' docs

makeUniFreqDicts dir = do
                     files <- getFiles [dir] ".tok"
                     mapM_ (\file -> do  uni <- readDocument file
                                         writeDict (file++".uni.freq") uni 1
                           )
                           files
                     putStrLn "Done: .uni.freq dictionaries"


makeBiFreqDicts dir = do
                     files <- getFiles [dir] ".tok"
                     mapM_ (\file -> do bi <- readDocumentBi file
                                        writeDictBi (file++".bi.freq") bi 1
                           )
                           files
                     putStrLn "Done: .bi.freq dictionaries"

makeSegFreqDicts dir = do
                     files <- getFiles [dir] ".seg"
                     mapM_ (\file -> do uni <- readDocument file
                                        writeDict (file++".freq") uni 1
                           )
                           files
                     putStrLn "Done: .seg.freq dictionaries"



mergeF [] [] fout = return ()
mergeF [] xs fout = BS.hPutStr fout (BS.unlines xs)
mergeF xs [] fout = BS.hPutStr fout (BS.unlines xs)
mergeF xss@(x:xs) yss@(y:ys) fout =
                               if (w1 == w2) 
                                  then do
                                         BS.hPutStrLn fout $ BS.concat [w1,BS.pack "\t",BS.pack $ show(freq1 + freq2)]
                                         mergeF xs ys fout
                                  else if (w1<w2)
                                          then do
                                                 BS.hPutStrLn fout x
                                                 mergeF xs yss fout
                                          else do
                                                 BS.hPutStrLn fout $ y
                                                 mergeF xss ys fout
                               where
                                 (w1,fr1) = BS.break (== '\t') x
                                 (w2,fr2) = BS.break (== '\t') y
                                 freq1 = read (BS.unpack fr1)::Int
                                 freq2 = read (BS.unpack fr2)::Int


mergeDictsT [] init  = return []
mergeDictsT [f] init = if init 
                              then do
                                   copyFile f $ f++"_"
                                   return [f++"_"]
                              else return [f]

mergeDictsT (f1:f2:fs) init = do
                          fs2  <- mergeDictsT fs init
                          fin1 <- BS.readFile f1
                          fin2 <- BS.readFile f2
                          (docs1:ws1) <- return (BS.lines fin1)
                          (docs2:ws2) <- return (BS.lines fin2)
                          fout <- IO.openFile (f1++"_")  IO.WriteMode
                          BS.hPutStrLn fout $ BS.pack $ show ((read (BS.unpack docs1)::Int) + (read (BS.unpack docs2)::Int )) 
                          mergeF ws1 ws2 fout
                          if (not init)
                            then do
                                   removeFile f1
                                   removeFile f2
                            else return ()
                          IO.hClose fout
                          return ((f1++"_"):fs2)

mergeDicts [] init = return []                          
mergeDicts fs init = do
                  fs2 <- mergeDictsT fs init
                  case (length fs2) of
                      1 -> return fs2
                      otherwise -> mergeDicts fs2 False

makeIDF :: [String] -> IO [String]
makeIDF [] = return []
makeIDF (f:fs) = do
                   fs1 <- makeIDF fs
                   fin <- BS.readFile f
                   (docs:ws) <- return (BS.lines fin)
                   fout <- IO.openFile (f++"_")  IO.WriteMode
                   BS.hPutStrLn fout $ docs
                   BS.hPutStr fout $ BS.unlines $ (map (\ line -> BS.append (BS.takeWhile (/='\t') line) $ BS.pack "\t1" ) ws) 
                   IO.hClose fout
                   return ((f++"_"):fs1)







buildIndex dir ext = do
                     files             <- getFiles [dir] ext
                     fs2               <- mergeIndInit files
                     mergeInd fs2


mergeIndInit [] = return []
mergeIndInit (f:fs) = do
                          fs2 <- mergeIndInit fs
                          fin <- BS.readFile f
                          (docs:ws) <- return (BS.lines fin)
                          idf  <-  return [ (w,(read (BS.unpack idf)::Double)) | x<-ws, let (w,idf) = BS.span (/='\t') x]
                          let idf'' = map (snd) idf
                          let len   = toEnum (length (idf''))::Double
                          let avg   = (sum idf'') / len
                          let idf''  = filter (\(k,v) -> v>=avg) idf
                          let idf2  = if (length idf'' > 400) then (DL.sort $ take 400  $ DL.sortBy (compare `on` snd) idf'') else idf'' 
                          BS.writeFile (f++"_") $ BS.unlines $ map (\(w,idf) -> BS.concat [w, BS.pack "\t", fName, BS.pack "\t",BS.pack (show idf)]) idf2
                          return ((f++"_"):fs2)
                        where 
                          fName = BS.pack $ dropExt ".txt.tok" f
--                          fName = BS.pack $ takeWhile (/='.') f
--                          fName = BS.pack $ takeWhile (/='.') $ spanEnd (/='/') f



dropExt sep [] = []
dropExt sep xs = if null end
                    then rest
                    else if DL.isPrefixOf sep end
                            then rest
                            else rest ++ ('.':(dropExt sep $ tail end))
                 where
                   (rest,end) = span (/='.') xs


spanEnd sep [] = []
spanEnd sep xs = if null end
                   then rest
                   else spanEnd sep $ tail end
                 where
                   (rest,end) = span sep xs

mergeInd [] = return []
mergeInd [fs] = return [fs]
mergeInd fs = do
                  fs2 <- mergeIndT fs
                  mergeInd fs2


mergeIndF [] [] fout = return ()
mergeIndF [] xs fout = BS.hPutStr fout (BS.unlines xs)
mergeIndF xs [] fout = BS.hPutStr fout (BS.unlines xs)
mergeIndF xss@(x:xs) yss@(y:ys) fout =
                               if (w1 == w2) 
                                  then do
                                         BS.hPutStrLn fout $ BS.concat [w1, fr1, fr2]
                                         mergeIndF xs ys fout
                                  else if (w1<w2)
                                          then do
                                                 BS.hPutStrLn fout x
                                                 mergeIndF xs yss fout
                                          else do
                                                 BS.hPutStrLn fout y
                                                 mergeIndF xss ys fout
                               where
                                 (w1,fr1) = BS.break (== '\t') x
                                 (w2,fr2) = BS.break (== '\t') y


mergeIndT [] = return []
mergeIndT [f]= return [f]
mergeIndT (f1:f2:fs) = do
                          fin1 <- BS.readFile f1
                          fin2 <- BS.readFile f2
                          ws1 <- return (BS.lines fin1)
                          ws2 <- return (BS.lines fin2)
                          fout <- IO.openFile (f1++"_")  IO.WriteMode
                          mergeIndF ws1 ws2 fout
                          IO.hClose fout
                          removeFile f1
                          removeFile f2
                          fs2  <- mergeIndT fs
                          return ((f1++"_"):fs2)
                       


trainAllCorpus corpusPath ext threshold classified = do
         tokeniseFiles     corpusPath ext classified
         makeUniFreqDicts  corpusPath
         makeBiFreqDicts   corpusPath
         files             <- getFiles [corpusPath] ".uni.freq"
         rezFile           <- mergeDicts files True
         renameFile        (rezFile!!0) ("dict.uni")
         files2            <- makeIDF files
         rezFile           <- mergeDicts files2 False
         renameFile        (rezFile!!0) ("dict.uniD")
         files             <- getFiles [corpusPath] ".bi.freq"
         rezFile           <- mergeDicts files True
         renameFile        (rezFile!!0) ("dict.bi")
         files2            <- makeIDF files
         rezFile           <- mergeDicts files2 False
         renameFile        (rezFile!!0) ("dict.biD")
         doDirSegmentation corpusPath ".tok" ("dict.uni") ("dict.bi") threshold
         makeSegFreqDicts  corpusPath
         files             <- getFiles [corpusPath] ".seg.freq"
         rezFile           <- mergeDicts files True
         renameFile        (rezFile!!0) ("dict.seg")
         files2            <- makeIDF files
         rezFile           <- mergeDicts files2 False
         renameFile        (rezFile!!0) ("dict.segD")


trainSegmentationCorpus corpusPath ext threshold = do
         doDirSegmentation corpusPath ".tok" ("dict.uni") ("dict.bi") threshold
         makeSegFreqDicts  corpusPath
         files             <- getFiles [corpusPath] ".seg.freq"
         rezFile           <- mergeDicts files True
         renameFile        (rezFile!!0) ("dict.seg")
         files2            <- makeIDF files
         rezFile           <- mergeDicts files2 False
         renameFile        (rezFile!!0) ("dict.segD")


makeSegmentationIndices corpusPath = do
         rezFile           <-  buildIndex        corpusPath ".seg.freq.tfidf"
         renameFile        (rezFile!!0) ("ind.seg")

makeIndices corpusPath = do
         rezFile           <-  buildIndex        corpusPath ".seg.freq.tfidf"
         renameFile        (rezFile!!0) ("ind.seg")
         rezFile           <-  buildIndex        corpusPath ".bi.freq.tfidf"
         renameFile        (rezFile!!0) ("ind.bi")
         rezFile           <-  buildIndex        corpusPath ".uni.freq.tfidf"
         renameFile        (rezFile!!0) ("ind.uni")
