-- Copyright   :  (c) Vidas Daudaravicius 2012
-- License     :  BSD3-style (see LICENSE)
-- Version     : 0.1

module Profiles where


import Files
import qualified Data.ByteString.Char8 as BS
import Data.List (groupBy, sortBy,foldl')
import qualified Data.Map as M
import qualified System.IO as IO
import Control.Monad (foldM)
import Ngram
import Dendogram (mergeSorted,cosine,preprocesBiFile,preprocesUniFile,preprocesSegFile)
import MakeDict
import DeepSeq


trainUniProfiles pDir = do
        files <- getFiles [pDir] ".txt"
        prof <- getProfileFiles files
        putStrLn $ "getProfileFiles: numberOfProfiles " ++ (show $ length prof)
        makeProfileFiles pDir ".tok.uni.freq" prof
        makeProfilesTFIDF pDir ("uni.profD") ".uni.freq.prof"
        putStrLn "done: makeProfilesTFIDF"


trainBiProfiles pDir = do
        files <- getFiles [pDir] ".txt"
        prof <- getProfileFiles files
        putStrLn $ "getProfileFiles: numberOfProfiles " ++ (show $ length prof)
        makeProfileFiles pDir ".tok.bi.freq" prof
        makeProfilesTFIDF pDir ("bi.profD") ".bi.freq.prof"
        putStrLn "done: makeProfilesTFIDF"

trainSegProfiles pDir = do
        files <- getFiles [pDir] ".txt"
        prof <- getProfileFiles files
        putStrLn $ "getProfileFiles: numberOfProfiles " ++ (show $ length prof)
        makeProfileFiles pDir ".tok.seg.freq" prof
        makeProfilesTFIDF pDir ("seg.profD") ".seg.freq.prof"
        putStrLn "done: makeProfilesTFIDF"

assignUniProfilesToFile pDir file rezCount classified = do
        preprocesUniFile file classified
        assignProfiles ("uniTest.txt") [file] pDir ".uni.freq.prof.ptfidf.bin" ".tok.uni.freq.tfidf"  rezCount classified
        putStrLn "done: assignProfiles"

assignBiProfilesToFile pDir file rezCount classified = do
        preprocesBiFile file classified
        assignProfiles ("biTest.txt")  [file] pDir ".bi.freq.prof.ptfidf.bin"  ".tok.bi.freq.tfidf" rezCount classified
        putStrLn "done: assignProfiles"

assignSegProfilesToFile pDir file threshold rezCount classified = do
        preprocesSegFile file threshold classified
        assignProfiles ("segTest.txt") [file] pDir ".seg.freq.prof.ptfidf.bin" ".tok.seg.freq.tfidf" rezCount  classified
        putStrLn "done: assignProfiles"


getProfiles file = do
                    h <- IO.openFile file IO.ReadMode
                    line <- IO.hGetLine h
                    IO.hClose h
                    return $!! map (\ p -> (p,file)) $ words line


getProfileFiles files = do
                       prf <- mapM getProfiles files
                       let prs2 = groupBy (\(a,_) (b,_) -> a == b)
                                          $ sortBy (\(a,_) (b,_) -> compare a b) 
                                          $ concat prf
                       return $!! [ ((fst $ head pr), (map snd pr)) | pr <- prs2 ]

makeProfile dir ext (pName,files) = do
                            dict <- foldM (\d f -> do
                                                     d2 <- readDict'' (f++ext)
                                                     return $ seq d mergeSorted d d2) [] files
                            hFile <- IO.openFile (dir++"/"++pName ++ ext ++".prof") IO.WriteMode
                            BS.hPutStrLn hFile $ BS.pack $ show 1
                            printDict hFile 1 $! dict
                            IO.hClose hFile
                               


makeProfileFiles dir ext prof = mapM_ (makeProfile dir ext) prof


makeProfilesTFIDF dir indF ext = do
                       files <- getFiles [dir] ext
                       dictD <- foldM (\ d f -> do
                                                 d2 <- readDict'' f
                                                 return $!! mergeSorted (map (\(w,fr) -> (w,1)) d2) d
                                      ) [] files
                       hFile <- IO.openFile indF IO.WriteMode
                       BS.hPutStrLn hFile $ BS.pack $ show 1
                       printDict hFile 1 $! (dictD::[(BS.ByteString, Double)])
                       IO.hClose hFile
                       avgDocLen <- avgDocLen files
                       putStrLn $ "done: the average length of profiles = " ++ (show avgDocLen)
                       dictD <- readDict''  indF
                       foldM (\ counter f -> do
                                       countTFIDF (M.fromAscList dictD) (toEnum $ length files) avgDocLen ".ptfidf" f
                                       fromDictToBinDict (f++".ptfidf")
                                       removeFile' f
                                       return $ counter+1 
                                       ) 0 files

getFileDescriptors file = do
                           hIn <- IO.openFile file IO.ReadMode
                           ln <- IO.hGetLine hIn
                           IO.hClose hIn
                           return $!! words ln

assignProfiles resFile testFiles dir profExt ext rezCount classified = do
                       prof <- getFiles [dir] profExt
                       hOut <- IO.openFile resFile IO.WriteMode
                       foldM (\ counter file -> do
                                       dict1 <- readDict'' $ file++ext
                                       putStrLn $ "profiles for: " ++ file++ext 
                                       desc <- getFileDescriptors file 
                                       min1 <- return $ foldl' (\ v1 (_,v2)-> min v1 v2) 1 dict1
                                       docT <- BS.readFile file
                                       let docLen = length $ BS.words docT
                                       let fn = (getFileName file)
                                       ls <- mapM (\ profF -> do
                                                         dict2 <- readBinDict profF
                                                         minT <- return $ foldl' (\ v1 (_,v2)-> min v1 v2) min1 dict1
                                                         return $!! (profF,(cosine (minT/10) dict1 dict2))
                                                      ) prof
                                       rs <- return  $!! take (if (not classified) then rezCount else length desc) $ sortBy (\(_,v1)(_,v2) -> compare v2 v1) ls
                                       IO.hPutStr hOut $ unlines 
                                                    $ map ( \(d,v) -> 
                                                            let fd = (getFileName d) 
                                                            in  fn ++ "\t" ++ fd 
                                                                   ++ "\t" ++ (show v) 
                                                                   ++ "\t" ++ (if (not classified) then "" else show $ elem fd desc)
--                                                                   ++ "\t" ++ (show $ length dict1)
--                                                                   ++ "\t" ++ (show docLen)
                                                          ) rs 
                                       IO.hFlush hOut
                                       return $ counter + 1
                             ) 0  testFiles
                       IO.hClose hOut
                       txt <- BS.readFile resFile
                       BS.putStrLn txt
                  
