-- Copyright   :  (c) Vidas Daudaravicius 2012
-- License     :  BSD3-style (see LICENSE)
-- Version     : 0.1

module Segmentation where

import qualified System.IO as IO
import qualified Data.Map as M
import Text
import Files
import Ngram
import Data.List (foldl')
import DeepSeq
import qualified Data.ByteString.Char8 as BS


collocabilityFunction = dice

------------------------------------------------------
------------------------------------------------------
---------------   Tokenisation   ---------------------
------------------------------------------------------
------------------------------------------------------
tokeniseFiles dir ext classified = do
                              fs <- getFiles [dir] ext
                              mapM_ (\f -> tokeniseFile f classified) fs
------------------------------------------------------
tokeniseFile fileName classified = do 
                              hFile1 <- IO.openFile fileName IO.ReadMode 
                              hFile2 <- IO.openFile (fileName++".tok") IO.WriteMode
--                              hSetEncoding hFile1 utf8
--                              hSetEncoding hFile2 utf8
                              if (classified)
                                then IO.hGetLine hFile1        -- skip descriptors, first line
                                else return ("")
                              tokeniseFile' hFile1 hFile2
                              IO.hClose hFile1
                              IO.hClose hFile2
--                              putStrLn ("tok: "++fileName++".tok")

tokeniseFile' hFile1 hFile2 = do
                                  isEnd <-IO.hIsEOF hFile1
                                  if(isEnd)
                                     then return ()
                                     else do
                                           line <- IO.hGetLine hFile1
                                           IO.hPutStrLn hFile2 (unwords $ tokenize line)
                                           tokeniseFile' hFile1 hFile2


------------------------------------------------------
------------------------------------------------------
---------------  Segmentation   ----------------------
------------------------------------------------------
------------------------------------------------------
--doFileSegmentation:: String -> String -> String -> IO()
doFileSegmentation file uniFile biFile threshold = do
               uni' <- readDict'' uniFile
               uni <- return $!! M.fromAscList uni'
               bi'  <- readDict'' biFile
               bi  <- return $!! M.fromAscList bi'
               corpusSize <- return $!! (foldl' (\y x -> y + (snd x)) 0 uni')
               segmentDocument uni bi corpusSize threshold file
------------------------------------------------------
--doDirSegmentation:: String -> String -> String -> String -> IO()
doDirSegmentation dir ext uniFile biFile threshold = do
                                                files <- getFiles [dir] ext
                                                uni' <- readDict'' uniFile
                                                uni <- return $!! M.fromAscList uni'
--                                                uni <- readDict'''' uniFile
                                                bi'  <- readDict'' biFile
                                                bi  <- return $!! M.fromAscList bi'
--                                                bi  <- readDict'''' biFile
                                                corpusSize <- return $! foldl' (+) 0 $ M.elems uni
                                                mapM_ (segmentDocument uni bi corpusSize threshold) files
------------------------------------------------------
segmentDocument uni bi corpusSize threshold fileName = do
                           text <- BS.readFile(fileName)
                           BS.writeFile (fileName++".seg") $ (BS.unlines $ makeTextSegmentation (BS.lines text) uni bi corpusSize threshold)
------------------------------------------------------
makeTextSegmentation lines uni bi corpusSize threshold = map (segLn) lines
                                           where
                                             segLn line  = (BS.unwords) . init . tail $ makeLineSegmentation ( [BS.pack "%."] ++ (BS.words line) ++ [BS.pack ".%"] ) uni bi corpusSize threshold
------------------------------------------------------
makeLineSegmentation ws uni bi corpusSize threshold' = setBoundaries threshold $! ws'
                                             where
                                               ws' = countCollocability ws uni bi corpusSize
                                               ini = map (snd) $! init ws'
                                               minV = minimum ini
                                               maxV = maximum ini
                                               avg = ( sum ini )/(fromIntegral (length ini))
                                               threshold = if threshold'<= 1 
                                                              then (avg - minV)*threshold' + minV
                                                              else maxV - (maxV - avg)*(2-threshold')
------------------------------------------------------
--countCollocability :: [String] -> M.Map String Double -> M.Map String Double -> Double -> [(String,Double)]
countCollocability []         uni bi corpusSize = []
countCollocability [w1]       uni bi corpusSize = [(w1,0)]
countCollocability (w1:w2:ws) uni bi corpusSize = (w1,(collocabilityFunction w1 w2 uni bi corpusSize)):(countCollocability (w2:ws) uni bi corpusSize)
------------------------------------------------------
setBoundaries threshold [(w1,c1),(w2,c2)] = [w1,w2]
setBoundaries threshold [(w1,c1),(w2,c2),(w3,c3)] = [w1,w2,w3]
setBoundaries threshold ((w1,c1):(w2,c2):(w3,c3):ws)
                       | (c1 + c3)/2 < c2 && c2>=threshold = setBoundaries threshold ((w1,c2):((BS.concat [w2, BS.pack "_", w3]),c3):ws)
                       | otherwise                         = w1:setBoundaries threshold ((w2,c2):(w3,c3):ws)
setBoundaries threshold _ = []
------------------------------------------------------
dice x y uni bi corpusSize = 2*fxy/(fx+fy)
                           where 
                  fxy = M.findWithDefault 0.1 (BS.concat [x, BS.pack " ", y]) bi 
                  fx  = M.findWithDefault 0.1 x               uni
                  fy  = M.findWithDefault 0.1 y               uni
------------------------------------------------------
logLikelihood x y uni bi corpusSize
         | fx==1 && fxy==1 = 0
         | fy==1 && fxy==1 = 0
         | otherwise       = 2*(fx-fxy)*log((fx-fxy)/(corpusSize-fxy)) - 2*fx*log(fx/corpusSize) + 2*log((fxy-fx)*log(fxy/fx))
                             *
                             2*(fy-fxy)*log((fy-fxy)/(corpusSize-fxy)) - 2*fy*log(fy/corpusSize) + 2*log((fxy-fy)*log(fxy/fy))
               where 
                  fxy = M.findWithDefault 0 (x ++ " " ++ y) bi 
                  fx  = M.findWithDefault 0 x               uni
                  fy  = M.findWithDefault 0 y               uni
------------------------------------------------------
