-- Copyright   :  (c) Vidas Daudaravicius 2012
-- License     :  BSD3-style (see LICENSE)
-- Version     : 0.1


import Dendogram
import MakeDict
import qualified System.IO as IO
import System.Environment
import Files
import Context
import Profiles
import qualified Data.ByteString.Char8 as BS
import Segmentation

makeSegmentationOnly fs = elem "makeSegmentationOnly" fs
isClassified fs = elem "classified" fs

main = do
         f <- getArgs
         let corpus = "corpus"
         let tokens = "tokens"
         let segments = "segments"
         if (length f == 0)
          then showHelp
          else
           case (f!!0) of
             "-trainUniProfiles"         -> do trainUniProfiles corpus
             "-trainBiProfiles"          -> do trainBiProfiles corpus
             "-trainSegProfiles"         -> do trainSegProfiles corpus
             "-assignUniProfilesToFile"  -> do assignUniProfilesToFile corpus (f!!1) (read (f!!2)::Int) (isClassified f)
             "-assignBiProfilesToFile"   -> do assignBiProfilesToFile corpus (f!!1) (read (f!!2)::Int) (isClassified f)
             "-assignSegProfilesToFile"  -> do assignSegProfilesToFile corpus (f!!1) ((read (f!!2)::Double)/100) (read (f!!3)::Int) (isClassified f)   -- threshold
             "-train"                    -> do trainAllCorpus corpus ".txt" ((read (f!!1)::Double)/100) (isClassified f)   -- threshold
                                               if(makeSegmentationOnly f)
                                                   then do
                                                         removeFiles [corpus] ".tok"
                                                         removeFiles [corpus] ".freq"
                                                   else do
                                                         countDirTFIDF corpus
                                                         makeIndices corpus
             "-fileSegmentation"         -> do tokeniseFile (f!!1) (isClassified f)
                                               doFileSegmentation ((f!!1)++".tok") "dict.uni" "dict.bi" ((read (f!!2)::Double)/100)   -- threshold
             "-directorySegmentation"    -> do tokeniseFiles (f!!1) ".txt" (isClassified f)
                                               doDirSegmentation (f!!1) ".tok" ("dict.uni") ("dict.bi") ((read (f!!2)::Double)/100)
             "-trainSegmentation"        -> do trainSegmentationCorpus corpus "" ((read (f!!1)::Double)/100)   -- threshold
                                               countSegmentationDirTFIDF corpus
                                               makeSegmentationIndices corpus
             "-lookForUniDocument"       -> lookForDocument  (f!!1) "uni" 0 (isClassified f) (read (f!!2)::Int) 
             "-lookForBiDocument"        -> lookForDocument  (f!!1) "bi" 0 (isClassified f) (read (f!!2)::Int)
             "-lookForSegDocument"       -> lookForDocument  (f!!1) "seg" ((read (f!!2)::Double)/100) (isClassified f)  (read (f!!3)::Int) 
             "-clean"                    -> do removeFiles [corpus] ".tok"
                                               removeFiles [corpus] ".seg"
                                               removeFiles [corpus] ".freq"
                                               removeFiles [corpus] ".tfidf"
                                               removeFiles [corpus] ".con"
                                               removeFiles [corpus] ".prof"
                                               removeFiles [corpus] ".ptfidf"
                                               removeFiles [corpus] ".bin"
                                               removeFiles [corpus] "_"
                                               removeFiles [""] ".avglen"
                                               removeFiles [""] ".profD"
                                               removeFiles [""] ".freq"
                                               removeFiles [""] ".tfidf"
                                               removeFiles [""] ".tok"
                                               removeFiles [""] ".seg"
--                                           removeFiles [""] "Test.txt"
--                                           removeFiles [""] ".profD"
--                                           removeFiles [""] ".uni"
--                                           removeFiles [""] ".seg"
--                                           removeFiles [""] ".tokD"
                                               removeFile' $ "bi.avglen"
                                               removeFile' $ "uni.avglen"
                                               removeFile' $ "seg.avglen"
                                               removeFile' $ "bi.profD"
                                               removeFile' $ "uni.profD"
                                               removeFile' $ "seg.profD"
                                               removeFile' $ "ind.uni"
                                               removeFile' $ "ind.bi"
                                               removeFile' $ "ind.seg"
                                               removeFile' $ "dict.tokD"
                                               removeFile' $ "dict.biD"
                                               removeFile' $ "dict.segD"
                                               removeFile' $ "dict.uniD"
                                               removeFile' $ "dict.uni"
                                               removeFile' $ "dict.bi"
                                               removeFile' $ "dict.seg"
                                               removeFile' $ "context.tokD"
                                               removeFile' $ "context.segD"
                                               removeFile' $ "context.ind.seg"
                                               removeFile' $ "context.ind.tok"
                                               removeFile' $ "context.seg.idf"
                                               removeFile' $ "context.tok.idf"
                                               removeFile' $ "segTest.txt"
                                               removeFile' $ "uniTest.txt"
                                               removeFile' $ "biTest.txt"
             "-show"                     -> do txt <- IO.readFile (f!!1)
                                               IO.putStrLn $ unlines $ take (read (f!!2)::Int) $ lines txt
             "-trainContextTokens"       -> makeWordContext corpus tokens ".tok" "context.tokD" (read (f!!1)::Int) "context.ind.tok" "context.tok.idf"
             "-trainContextSegments"     -> makeWordContext corpus segments ".tok.seg" "context.segD" (read (f!!1)::Int) "context.ind.seg" "context.seg.idf"
             "-lookForTokens"            -> look "context.ind.tok" "context.tok.idf"  (f!!1) (read (f!!2)::Int) 
             "-lookForSegments"          -> look "context.ind.seg" "context.seg.idf"  (f!!1) (read (f!!2)::Int)
             "-partOfSegment"            -> do hIn <- IO.openFile "context.ind.seg" IO.ReadMode
                                               cs <- partOfSegment hIn $ BS.pack $ "_" ++ (f!!1) ++ "_"
                                               IO.hClose hIn
                                               BS.putStrLn $ BS.unlines cs
             "-partOfToken"              -> do hIn <- IO.openFile "context.ind.tok" IO.ReadMode
                                               cs <- partOfToken hIn $ BS.pack (f!!1)
                                               IO.hClose hIn
                                               BS.putStrLn $ BS.unlines cs
             _                           ->  showHelp

showHelp = do t <- IO.readFile "readme.txt"
              IO.putStrLn t

